// Copyright © 2019 The Tekton Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package task

import (
	"errors"
	"fmt"
	"strings"
	"testing"
	"time"

	"gotest.tools/v3/golden"
	"k8s.io/apimachinery/pkg/runtime"

	"github.com/jonboulle/clockwork"
	"github.com/tektoncd/cli/pkg/test"
	cb "github.com/tektoncd/cli/pkg/test/builder"
	"github.com/tektoncd/pipeline/pkg/apis/pipeline/v1alpha1"
	"github.com/tektoncd/pipeline/pkg/reconciler/pipelinerun/resources"
	pipelinetest "github.com/tektoncd/pipeline/test"
	tb "github.com/tektoncd/pipeline/test/builder"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	k8stest "k8s.io/client-go/testing"
	"knative.dev/pkg/apis"
)

func TestTaskDescribe_Invalid_Namespace(t *testing.T) {
	cs, _ := test.SeedTestData(t, pipelinetest.Data{})
	p := &test.Params{Tekton: cs.Pipeline, Kube: cs.Kube}

	task := Command(p)
	_, err := test.ExecuteCommand(task, "desc", "bar", "-n", "invalid")
	if err == nil {
		t.Errorf("Error expected here")
	}
	expected := "namespaces \"invalid\" not found"
	test.AssertOutput(t, expected, err.Error())
}

func TestTaskDescribe_Empty(t *testing.T) {
	cs, _ := test.SeedTestData(t, pipelinetest.Data{
		Namespaces: []*corev1.Namespace{
			{
				ObjectMeta: metav1.ObjectMeta{
					Name: "ns",
				},
			},
		},
	})
	p := &test.Params{Tekton: cs.Pipeline, Kube: cs.Kube}

	task := Command(p)
	_, err := test.ExecuteCommand(task, "desc", "bar", "-n", "ns")
	if err == nil {
		t.Errorf("Error expected here")
	}
	expected := "tasks.tekton.dev \"bar\" not found"
	test.AssertOutput(t, expected, err.Error())
}

func TestTaskDescribe_OnlyName(t *testing.T) {
	cs, _ := test.SeedTestData(t, pipelinetest.Data{
		Tasks: []*v1alpha1.Task{
			tb.Task("task-1", "ns"),
			tb.Task("task", "ns-2"),
		},
		Namespaces: []*corev1.Namespace{
			{
				ObjectMeta: metav1.ObjectMeta{
					Name: "ns",
				},
			},
			{
				ObjectMeta: metav1.ObjectMeta{
					Name: "ns-2",
				},
			},
		},
	})

	p := &test.Params{Tekton: cs.Pipeline, Kube: cs.Kube}
	p.SetNamespace("ns")
	task := Command(p)
	out, err := test.ExecuteCommand(task, "desc", "task-1")
	if err != nil {
		t.Errorf("Unexpected error: %v", err)
	}
	golden.Assert(t, out, fmt.Sprintf("%s.golden", t.Name()))
}

func TestTaskDescribe_OnlyNameDiffNameSpace(t *testing.T) {
	cs, _ := test.SeedTestData(t, pipelinetest.Data{
		Tasks: []*v1alpha1.Task{
			tb.Task("task-1", "ns"),
			tb.Task("task", "ns-2"),
		},
		Namespaces: []*corev1.Namespace{
			{
				ObjectMeta: metav1.ObjectMeta{
					Name: "ns",
				},
			},
			{
				ObjectMeta: metav1.ObjectMeta{
					Name: "ns-2",
				},
			},
		},
	})

	p := &test.Params{Tekton: cs.Pipeline, Kube: cs.Kube}
	p.SetNamespace("ns")
	task := Command(p)
	out, err := test.ExecuteCommand(task, "desc", "task", "-n", "ns-2")
	if err != nil {
		t.Errorf("Unexpected error: %v", err)
	}
	golden.Assert(t, out, fmt.Sprintf("%s.golden", t.Name()))
}

func TestTaskDescribe_OnlyNameParams(t *testing.T) {
	cs, _ := test.SeedTestData(t, pipelinetest.Data{
		Tasks: []*v1alpha1.Task{
			tb.Task("task-1", "ns",
				tb.TaskSpec(
					tb.TaskInputs(
						tb.InputsParamSpec("myarg", v1alpha1.ParamTypeString),
						tb.InputsParamSpec("myprint", v1alpha1.ParamTypeString),
						tb.InputsParamSpec("myarray", v1alpha1.ParamTypeArray),
					),
				),
			),
			tb.Task("task", "ns-2"),
		},
		Namespaces: []*corev1.Namespace{
			{
				ObjectMeta: metav1.ObjectMeta{
					Name: "ns",
				},
			},
			{
				ObjectMeta: metav1.ObjectMeta{
					Name: "ns-2",
				},
			},
		},
	})

	p := &test.Params{Tekton: cs.Pipeline, Kube: cs.Kube}
	p.SetNamespace("ns")
	task := Command(p)
	out, err := test.ExecuteCommand(task, "desc", "task-1")
	if err != nil {
		t.Errorf("Unexpected error: %v", err)
	}
	golden.Assert(t, out, fmt.Sprintf("%s.golden", t.Name()))
}

func TestTaskDescribe_Full(t *testing.T) {
	clock := clockwork.NewFakeClock()
	cs, _ := test.SeedTestData(t, pipelinetest.Data{
		Tasks: []*v1alpha1.Task{
			tb.Task("task-1", "ns",
				tb.TaskSpec(
					tb.TaskInputs(
						tb.InputsResource("my-repo", v1alpha1.PipelineResourceTypeGit),
						tb.InputsResource("my-image", v1alpha1.PipelineResourceTypeImage),
						tb.InputsResource("source-repo", v1alpha1.PipelineResourceTypeGit),
						tb.InputsParamSpec("myarg", v1alpha1.ParamTypeString),
						tb.InputsParamSpec("myarray", v1alpha1.ParamTypeArray),
						tb.InputsParamSpec("print", v1alpha1.ParamTypeString, tb.ParamSpecDefault("somethingdifferent")),
						tb.InputsParamSpec("output", v1alpha1.ParamTypeArray, tb.ParamSpecDefault("booms", "booms", "booms")),
					),
					tb.TaskOutputs(
						tb.OutputsResource("code-image", v1alpha1.PipelineResourceTypeImage),
						tb.OutputsResource("artifact-image", v1alpha1.PipelineResourceTypeImage),
					),
					tb.Step("busybox",
						tb.StepName("hello"),
					),
					tb.Step("busybox",
						tb.StepName("exit"),
					),
				),
			),
		},
		TaskRuns: []*v1alpha1.TaskRun{
			tb.TaskRun("tr-1", "ns",
				tb.TaskRunLabel("tekton.dev/task", "task-1"),
				tb.TaskRunSpec(tb.TaskRunTaskRef("task-1")),
				tb.TaskRunStatus(
					tb.StatusCondition(apis.Condition{
						Status: corev1.ConditionFalse,
						Reason: resources.ReasonFailed,
					}),
					tb.TaskRunStartTime(clock.Now()),
					cb.TaskRunCompletionTime(clock.Now().Add(5*time.Minute)),
				),
			),
			tb.TaskRun("tr-2", "ns",
				tb.TaskRunLabel("tekton.dev/task", "task-1"),
				tb.TaskRunSpec(tb.TaskRunTaskRef("task-1")),
				tb.TaskRunStatus(
					tb.StatusCondition(apis.Condition{
						Status: corev1.ConditionTrue,
						Reason: resources.ReasonSucceeded,
					}),
					tb.TaskRunStartTime(clock.Now().Add(10*time.Minute)),
					cb.TaskRunCompletionTime(clock.Now().Add(17*time.Minute)),
				),
			),
		},
		Namespaces: []*corev1.Namespace{
			{
				ObjectMeta: metav1.ObjectMeta{
					Name: "ns",
				},
			},
		},
	})

	p := &test.Params{Tekton: cs.Pipeline, Clock: clock, Kube: cs.Kube}
	task := Command(p)
	clock.Advance(20 * time.Minute)
	out, err := test.ExecuteCommand(task, "desc", "task-1")
	if err != nil {
		t.Errorf("Unexpected error: %v", err)
	}
	golden.Assert(t, out, fmt.Sprintf("%s.golden", t.Name()))
}

func TestTaskDescribe_TaskRunError(t *testing.T) {
	cs, _ := test.SeedTestData(t, pipelinetest.Data{
		Tasks: []*v1alpha1.Task{
			tb.Task("task-1", "ns"),
			tb.Task("task", "ns-2"),
		},
		Namespaces: []*corev1.Namespace{
			{
				ObjectMeta: metav1.ObjectMeta{
					Name: "ns",
				},
			},
			{
				ObjectMeta: metav1.ObjectMeta{
					Name: "ns-2",
				},
			},
		},
	})

	p := &test.Params{Tekton: cs.Pipeline, Kube: cs.Kube}
	p.SetNamespace("ns")

	cs.Pipeline.PrependReactor("list", "taskruns", func(action k8stest.Action) (bool, runtime.Object, error) {
		return true, nil, errors.New("fake list taskrun error")
	})

	task := Command(p)
	out, err := test.ExecuteCommand(task, "desc", "task-1")
	if err == nil {
		t.Errorf("Expected error got nil")
	}
	expected := "failed to get taskruns for task task-1 \nError: fake list taskrun error\n"
	test.AssertOutput(t, expected, out)
	test.AssertOutput(t, "fake list taskrun error", err.Error())
}

func TestTaskDescribe_custom_output(t *testing.T) {
	name := "task"
	expected := "task.tekton.dev/" + name

	clock := clockwork.NewFakeClock()

	tasks := []*v1alpha1.Task{
		tb.Task(name, "ns"),
	}

	cs, _ := test.SeedTestData(t, pipelinetest.Data{
		Tasks: tasks,
		Namespaces: []*corev1.Namespace{
			{
				ObjectMeta: metav1.ObjectMeta{
					Name: "ns",
				},
			},
		},
	})

	p := &test.Params{Tekton: cs.Pipeline, Clock: clock, Kube: cs.Kube}
	task := Command(p)
	got, err := test.ExecuteCommand(task, "desc", "-o", "name", "-n", "ns", name)
	if err != nil {
		t.Errorf("Unexpected error: %v", err)
	}

	got = strings.TrimSpace(got)
	if got != expected {
		t.Errorf("Result should be '%s' != '%s'", got, expected)
	}
}
